#!/usr/bin/env python

import sys
import os
import codecs
from mako.template import Template
from mako.lookup import TemplateLookup

def render(hosts, vars={}, tpl_dirs=[]):
    if not os.path.isdir('cmdb'):
        os.mkdir('cmdb')

    lookup = TemplateLookup(directories=tpl_dirs,
                            default_filters=['decode.utf8'],
                            input_encoding='utf-8',
                            output_encoding='utf-8',
                            encoding_errors='replace')

    # Render host overview
    template = lookup.get_template('markdown_split_overview.tpl')
    out_file = os.path.join('cmdb', 'overview.md')
    output = template.render(hosts=hosts, **vars).lstrip().decode('utf8')
    with codecs.open(out_file, 'w', encoding='utf8') as f:
        f.write(output)

    # Render host details
    template = lookup.get_template('markdown_split_detail.tpl')
    for hostname, host in hosts.items():
        out_file = os.path.join('cmdb', '{0}.md'.format(hostname))
        output = template.render(hostname=hostname, host=host, **vars).lstrip().decode('utf8')
        with codecs.open(out_file, 'w', encoding='utf8') as f:
            f.write(output)



import sys
import os
import json
import subprocess
import codecs
import logging
from . import ihateyaml
import ansiblecmdb.util as util
import ansiblecmdb.parser as parser


def strip_exts(s, exts):
    """
    Given a string and an interable of extensions, strip the extenion off the
    string if the string ends with one of the extensions.
    """
    f_split = os.path.splitext(s)
    if f_split[1] in exts:
        return f_split[0]
    else:
        return s


class Ansible(object):
    """
    The Ansible class is responsible for gathering host information.
    """
    def __init__(self, fact_dirs, inventory_paths=None, fact_cache=False, debug=False):
        """
        `fact_dirs` is a list of paths to directories containing facts gathered
        by ansible's 'setup' module.

        `inventory_paths` is a list with files or directories containing the
        inventory. It will be scanned to extract groups, variables and
        additional facts. If entries point to a file, it's read as a
        hosts file. If it's a directory, it is scanned for hosts files and
        dynamic inventory scripts.
        """
        self.fact_dirs = fact_dirs
        if inventory_paths is None:
            self.inventory_paths = []
        else:
            self.inventory_paths = inventory_paths
        self.fact_cache = fact_cache  # fact dirs are fact-caches
        self.debug = debug
        self.hosts = {}
        self.log = logging.getLogger(__name__)

        # Process facts gathered by Ansible's setup module of fact caching.
        for fact_dir in self.fact_dirs:
            self._parse_fact_dir(fact_dir, self.fact_cache)

        # Scan the inventory for known hosts.
        for inventory_path in self.inventory_paths:
            self._handle_inventory(inventory_path)

        # Scan for host vars and apply them
        for inventory_path in self.inventory_paths:
            self._parse_hostvar_dir(inventory_path)

        # Scan for group vars and apply them
        for inventory_path in self.inventory_paths:
            self._parse_groupvar_dir(inventory_path)

    def _handle_inventory(self, inventory_path):
        """
        Scan inventory. As Ansible is a big mess without any kind of
        preconceived notion of design, there are several (and I use that word
        lightly) different ways inventory_path can be handled:

          - a non-executable file: handled as a Ansible 'hosts' file.
          - an executable file: handled as a dynamic inventory file.
          - a directory: scanned for Ansible 'hosts' and dynamic inventory
            files.
        """
        self.log.debug("Determining type of inventory_path {}".format(inventory_path))
        if os.path.isfile(inventory_path) and \
           util.is_executable(inventory_path):
            # It's a file and it's executable. Handle as dynamic inventory script
            self.log.debug("{} is a executable. Handle as dynamic inventory script".format(inventory_path))
            self._parse_dyn_inventory(inventory_path)
        elif os.path.isfile(inventory_path):
            # Static inventory hosts file
            self.log.debug("{} is a file. Handle as static inventory file".format(inventory_path))
            self._parse_hosts_inventory(inventory_path)
        elif os.path.isdir(inventory_path):
            # Directory
            self.log.debug("{} is a dir. Just try most files to see if they happen to be inventory files".format(inventory_path))

            # Don't parse folder as inventory if it is a .git or group/host_vars
            if any(os.path.basename(inventory_path) == name for name in ['.git', 'group_vars', 'host_vars']):
                return

            # Scan directory
            for fname in os.listdir(inventory_path):
                # Skip files that end with certain extensions or characters
                if any(fname.endswith(ext) for ext in ["~", ".orig", ".bak", ".ini", ".cfg", ".retry", ".pyc", ".pyo", ".gitignore"]):
                    continue

                self._handle_inventory(os.path.join(inventory_path, fname))
        else:
            raise IOError("Invalid inventory file / dir: '{0}'".format(inventory_path))

    def _parse_hosts_inventory(self, inventory_path):
        """
        Read all the available hosts inventory information into one big list
        and parse it.
        """
        hosts_contents = []
        if os.path.isdir(inventory_path):
            self.log.debug("Inventory path {} is a dir. Looking for inventory files in that dir.".format(inventory_path))
            for fname in os.listdir(inventory_path):
                # Skip .git folder
                if fname == '.git':
                    continue
                path = os.path.join(inventory_path, fname)
                if os.path.isdir(path):
                    continue
                with codecs.open(path, 'r', encoding='utf8') as f:
                    hosts_contents += f.readlines()
        else:
            self.log.debug("Inventory path {} is a file. Reading as inventory.".format(inventory_path))
            with codecs.open(inventory_path, 'r', encoding='utf8') as f:
                hosts_contents = f.readlines()

        # Parse inventory and apply it to the hosts
        hosts_parser = parser.HostsParser(hosts_contents)
        for hostname, key_values in hosts_parser.hosts.items():
            self.update_host(hostname, key_values)

    def _parse_hostvar_dir(self, inventory_path):
        """
        Parse host_vars dir, if it exists.
        """
        # inventory_path could point to a `hosts` file, or to a dir. So we
        # construct the location to the `host_vars` differently.
        if os.path.isdir(inventory_path):
            path = os.path.join(inventory_path, 'host_vars')
        else:
            path = os.path.join(os.path.dirname(inventory_path), 'host_vars')

        self.log.debug("Parsing host vars (dir): {0}".format(path))
        if not os.path.exists(path):
            self.log.info("No such dir {0}".format(path))
            return

        for entry in os.listdir(path):
            # Skip .git folder
            if entry == '.git':
                continue
            full_path = os.path.join(path, entry)

            # file or dir name is the hostname
            hostname = strip_exts(entry, ('.yml', '.yaml', '.json'))

            if os.path.isfile(full_path):
                # Parse contents of file as host vars.
                self._parse_hostvar_file(hostname, full_path)
            elif os.path.isdir(full_path):
                # Parse each file in the directory as a file containing
                # variables for the host.
                for file_entry in os.listdir(full_path):
                    p = os.path.join(full_path, file_entry)
                    self._parse_hostvar_file(hostname, p)

    def _parse_hostvar_file(self, hostname, path):
        """
        Parse a host var file and apply it to host `hostname`.
        """
        # Check for ansible-vault files, because they're valid yaml for
        # some reason... (psst, the reason is that yaml sucks)
        first_line = open(path, 'r').readline()
        if first_line.startswith('$ANSIBLE_VAULT'):
            self.log.warning("Skipping encrypted vault file {0}".format(path))
            return

        try:
            self.log.debug("Reading host vars from {}".format(path))
            f = codecs.open(path, 'r', encoding='utf8')
            invars = ihateyaml.safe_load(f)
            f.close()
        except Exception as err:
            # Just catch everything because yaml...
            self.log.warning("Yaml couldn't load '{0}'. Skipping. Error was: {1}".format(path, err))
            return

        if invars is None:
            # Empty file or whatever. This is *probably* a side-effect of our
            # own yaml.SafeLoader implementation ('ihateyaml'), because this
            # problem didn't exist before.
            return

        if hostname == "all":
            # Hostname 'all' is special and applies to all hosts
            for hostname in self.hosts_all():
                self.update_host(hostname, {'hostvars': invars}, overwrite=False)
        else:
            self.update_host(hostname, {'hostvars': invars}, overwrite=True)

    def _parse_groupvar_dir(self, inventory_path):
        """
        Parse group_vars dir, if it exists. Encrypted vault files are skipped.
        """
        # inventory_path could point to a `hosts` file, or to a dir. So we
        # construct the location to the `group_vars` differently.
        if os.path.isdir(inventory_path):
            path = os.path.join(inventory_path, 'group_vars')
        else:
            path = os.path.join(os.path.dirname(inventory_path), 'group_vars')

        self.log.debug("Parsing group vars (dir): {0}".format(path))
        if not os.path.exists(path):
            self.log.info("No such dir {0}".format(path))
            return

        for (dirpath, dirnames, filenames) in os.walk(path):
            for filename in filenames:
                full_path = os.path.join(dirpath, filename)

                # filename is the group name
                groupname = strip_exts(filename, ('.yml', '.yaml', '.json'))

                try:
                    self.log.debug("Reading group vars from {}".format(full_path))
                    f = codecs.open(full_path, 'r', encoding='utf8')
                    invars = ihateyaml.safe_load(f)
                    f.close()
                except Exception as err:
                    # Just catch everything because yaml...
                    self.log.warning("Yaml couldn't load '{0}' because '{1}'. Skipping".format(full_path, err))
                    continue  # Go to next file

                if groupname == 'all':
                    # groupname 'all' is special and applies to all hosts.
                    for hostname in self.hosts_all():
                        self.update_host(hostname, {'hostvars': invars}, overwrite=False)
                else:
                    for hostname in self.hosts_in_group(groupname):
                        self.update_host(hostname, {'hostvars': invars}, overwrite=False)

    def _parse_fact_dir(self, fact_dir, fact_cache=False):
        """
        Walk through a directory of JSON files and extract information from
        them. This is used for both the Ansible fact gathering (setup module)
        output and custom variables.
        """
        self.log.debug("Parsing fact dir: {0}".format(fact_dir))
        if not os.path.isdir(fact_dir):
            raise IOError("Not a directory: '{0}'".format(fact_dir))

        flist = []
        for (dirpath, dirnames, filenames) in os.walk(fact_dir):
            flist.extend(filenames)
            break

        for fname in flist:
            if fname.startswith('.'):
                continue
            self.log.debug("Reading host facts from {0}".format(os.path.join(fact_dir, fname)))
            hostname = fname

            fd = codecs.open(os.path.join(fact_dir, fname), 'r', encoding='utf8')
            s = fd.readlines()
            fd.close()
            try:
                x = json.loads(''.join(s))
                # for compatibility with fact_caching=jsonfile
                # which omits the "ansible_facts" parent key added by the setup module
                if fact_cache:
                    x = json.loads('{ "ansible_facts": ' + ''.join(s) + ' }')
                self.update_host(hostname, x)
                self.update_host(hostname, {'name': hostname})
            except ValueError as e:
                # Ignore non-JSON files (and bonus errors)
                self.log.warning("Error parsing: %s: %s" % (fname, e))

    def _parse_dyn_inventory(self, script):
        """
        Execute a dynamic inventory script and parse the results.
        """
        self.log.debug("Reading dynamic inventory {0}".format(script))
        try:
            proc = subprocess.Popen([script, '--list'],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    close_fds=True)
            stdout, stderr = proc.communicate(input)
            if proc.returncode != 0:
                sys.stderr.write("Dynamic inventory script '{0}' returned "
                                 "exitcode {1}\n".format(script,
                                                         proc.returncode))
                for line in stderr:
                    sys.stderr.write(line)

            dyninv_parser = parser.DynInvParser(stdout.decode('utf8'))
            for hostname, key_values in dyninv_parser.hosts.items():
                self.update_host(hostname, key_values)
        except OSError as err:
            sys.stderr.write("Exception while executing dynamic inventory script '{0}':\n\n".format(script))
            sys.stderr.write(str(err) + '\n')

    def update_host(self, hostname, key_values, overwrite=True):
        """
        Update a hosts information. This is called by various collectors such
        as the ansible setup module output and the hosts parser to add
        informatio to a host. It does some deep inspection to make sure nested
        information can be updated.
        """
        default_empty_host = {
            'name': hostname,
            'hostvars': {},
        }
        host_info = self.hosts.get(hostname, default_empty_host)
        util.deepupdate(host_info, key_values, overwrite=overwrite)
        self.hosts[hostname] = host_info

    def hosts_all(self):
        """
        Return a list of all hostnames.
        """
        return [hostname for hostname, hostinfo in self.hosts.items()]

    def hosts_in_group(self, groupname):
        """
        Return a list of hostnames that are in a group.
        """
        result = []
        for hostname, hostinfo in self.hosts.items():
            if 'groups' in hostinfo:
                if groupname in hostinfo['groups']:
                    result.append(hostname)
            else:
                hostinfo['groups'] = [groupname]
        return result



import os
import imp
from mako.template import Template
from mako.lookup import TemplateLookup


class Render:
    def __init__(self, tpl, tpl_dirs):
        self.tpl = tpl
        self.tpl_dirs = tpl_dirs
        self.tpl_possibilities = self._tpl_possibilities()
        self.tpl_file = self._find_tpl()

    def _tpl_possibilities(self):
        """
        Construct a list of possible paths to templates.
        """
        tpl_possibilities = [
            os.path.realpath(self.tpl)
        ]
        for tpl_dir in self.tpl_dirs:
            tpl_possibilities.append(os.path.realpath(os.path.join(tpl_dir, "{0}.tpl".format(self.tpl))))
            tpl_possibilities.append(os.path.realpath(os.path.join(tpl_dir, "{0}.py".format(self.tpl))))

        return tpl_possibilities

    def _find_tpl(self):
        """
        Find a template in the list of possible paths.
        """
        for tpl_possibility in self.tpl_possibilities:
            if os.path.isfile(tpl_possibility):
                return tpl_possibility

        return None

    def render(self, hosts, vars={}):
        """
        Render a mako or .py file.
        """
        if self.tpl_file.endswith(".tpl"):
            return self._render_mako(hosts, vars)
        elif self.tpl_file.endswith(".py"):
            return self._render_py(hosts, vars)
        else:
            raise ValueError("Don't know how to handle '{0}'".format(self.tpl_file))

    def _render_mako(self, hosts, vars={}):
        lookup = TemplateLookup(directories=self.tpl_dirs,
                                default_filters=['decode.utf8'],
                                input_encoding='utf-8',
                                output_encoding='utf-8',
                                encoding_errors='replace')
        template = Template(filename=self.tpl_file,
                            lookup=lookup,
                            default_filters=['decode.utf8'],
                            input_encoding='utf-8',
                            output_encoding='utf-8')
        return template.render(hosts=hosts, **vars)

    def _render_py(self, hosts, vars={}):
        module = imp.load_source('r', self.tpl_file)
        return module.render(hosts, vars=vars, tpl_dirs=self.tpl_dirs)



import sys
import unittest
import imp
import os

sys.path.insert(0, os.path.realpath('../lib'))
sys.path.insert(0, os.path.realpath('../src'))
import ansiblecmdb


class ExtendTestCase(unittest.TestCase):
    """
    Test the extending of facts.
    """
    def testExtendOverrideParams(self):
        """
        Test that we can override a native fact
        """
        fact_dirs = ['f_extend/out_setup', 'f_extend/extend']
        ansible = ansiblecmdb.Ansible(fact_dirs)
        env_editor = ansible.hosts['debian.dev.local']['ansible_facts']['ansible_env']['EDITOR']
        self.assertEqual(env_editor, 'nano')

    def testExtendAddParams(self):
        """
        Test that we can add new facts
        """
        fact_dirs = ['f_extend/out_setup', 'f_extend/extend']
        ansible = ansiblecmdb.Ansible(fact_dirs)
        software = ansible.hosts['debian.dev.local']['software']
        self.assertIn('Apache2', software)


class HostParseTestCase(unittest.TestCase):
    """
    Test specifics of the hosts inventory parser
    """
    def testChildGroupHosts(self):
        """
        Test that children groups contain all hosts they should.
        """
        fact_dirs = ['f_hostparse/out']
        inventories = ['f_hostparse/hosts']
        ansible = ansiblecmdb.Ansible(fact_dirs, inventories)
        groups = ansible.hosts['db.dev.local']['groups']
        self.assertIn('db', groups)
        self.assertIn('dev', groups)
        self.assertIn('dev_local', groups)

    def testChildGroupVars(self):
        """
        Test that all vars applied against a child group are set on the hosts.
        """
        fact_dirs = ['f_hostparse/out']
        inventories = ['f_hostparse/hosts']
        ansible = ansiblecmdb.Ansible(fact_dirs, inventories)
        host_vars = ansible.hosts['db.dev.local']['hostvars']
        self.assertEqual(host_vars['function'], 'db')
        self.assertEqual(host_vars['dtap'], 'dev')

    def testExpandHostDef(self):
        """
        Verify that host ranges are properly expanded. E.g. db[01-03].local ->
        db01.local, db02.local, db03.local.
        """
        fact_dirs = ['f_hostparse/out']
        inventories = ['f_hostparse/hosts']
        ansible = ansiblecmdb.Ansible(fact_dirs, inventories)
        self.assertIn('web02.dev.local', ansible.hosts)
        self.assertIn('fe03.dev02.local', ansible.hosts)


class InventoryTestCase(unittest.TestCase):
    def testHostsDir(self):
        """
        Verify that we can specify a directory as the hosts inventory file and
        that all files are parsed.
        """
        fact_dirs = ['f_inventory/out']
        inventories = ['f_inventory/hostsdir']
        ansible = ansiblecmdb.Ansible(fact_dirs, inventories)
        host_vars = ansible.hosts['db.dev.local']['hostvars']
        groups = ansible.hosts['db.dev.local']['groups']
        self.assertEqual(host_vars['function'], 'db')
        self.assertIn('db', groups)

    def testDynInv(self):
        """
        Verify that we can specify a path to a dynamic inventory as the
        inventory file, and it will be executed, it's output parsed and added
        as available hosts.
        """
        fact_dirs = ['f_inventory/out'] # Reuse f_hostparse
        inventories = ['f_inventory/dyninv.py']
        ansible = ansiblecmdb.Ansible(fact_dirs, inventories)
        self.assertIn('host5.example.com', ansible.hosts)
        host_vars = ansible.hosts['host5.example.com']['hostvars']
        groups = ansible.hosts['host5.example.com']['groups']
        self.assertEqual(host_vars['b'], False)
        self.assertIn("atlanta", groups)

    def testMixedDir(self):
        """
        Verify that a mixed dir of hosts files and dynamic inventory scripts is
        parsed correctly.
        """
        fact_dirs = ['f_inventory/out']
        inventories = ['f_inventory/mixeddir']
        ansible = ansiblecmdb.Ansible(fact_dirs, inventories)
        # results from dynamic inventory
        self.assertIn("host4.example.com", ansible.hosts)
        self.assertIn("moocow.example.com", ansible.hosts)
        # results from normal hosts file.
        self.assertIn("web03.dev.local", ansible.hosts)
        # INI file ignored.
        self.assertNotIn("ini_setting", ansible.hosts)


class FactCacheTestCase(unittest.TestCase):
    """
    Test that we properly read fact-cached output dirs.
    """
    def testFactCache(self):
        fact_dirs = ['f_factcache/out']
        inventories = ['f_factcache/hosts']
        ansible = ansiblecmdb.Ansible(fact_dirs, inventories, fact_cache=True)
        host_vars = ansible.hosts['debian.dev.local']['hostvars']
        groups = ansible.hosts['debian.dev.local']['groups']
        ansible_facts = ansible.hosts['debian.dev.local']['ansible_facts']
        self.assertIn('dev', groups)
        self.assertEqual(host_vars['dtap'], 'dev')
        self.assertIn('ansible_env', ansible_facts)


if __name__ == '__main__':
    unittest.main(exit=True)

    try:
        os.unlink('../src/ansible-cmdbc') # FIXME: Where is this coming from? Our weird import I assume.
    except Exception:
        pass



import copy
import os
import stat


def is_executable(path):
    """
    Determine whether `path` points to an executable file.
    """
    return stat.S_IXUSR & os.stat(path)[stat.ST_MODE]


def deepupdate(target, src, overwrite=True):
    """Deep update target list, dict or set or other iterable with src
    For each k,v in src: if k doesn't exist in target, it is deep copied from
    src to target. Otherwise, if v is a list, target[k] is extended with
    src[k]. If v is a set, target[k] is updated with v, If v is a dict,
    recursively deep-update it. If `overwrite` is False, existing values in
    target will not be overwritten.

    Examples:
    >>> t = {'name': 'Ferry', 'hobbies': ['programming', 'sci-fi']}
    >>> deepupdate(t, {'hobbies': ['gaming']})
    >>> print t
    {'name': 'Ferry', 'hobbies': ['programming', 'sci-fi', 'gaming']}
    """
    for k, v in src.items():
        if type(v) == list:
            if not k in target:
                target[k] = copy.deepcopy(v)
            elif overwrite is True:
                target[k].extend(v)
        elif type(v) == dict:
            if not k in target:
                target[k] = copy.deepcopy(v)
            else:
                deepupdate(target[k], v, overwrite=overwrite)
        elif type(v) == set:
            if not k in target:
                target[k] = v.copy()
            elif overwrite is True:
                target[k].update(v.copy())
        else:
            if k not in target or overwrite is True:
                target[k] = copy.copy(v)


def find_path(dirs, path_to_find):
    """
    Go through a bunch of dirs and see if dir+path_to_find exists there.
    Returns the first dir that matches. Otherwise, return None.
    """
    for dir in dirs:
        if os.path.exists(os.path.join(dir, path_to_find)):
            return dir
    return None


def to_bool(s):
    """
    Convert string `s` into a boolean. `s` can be 'true', 'True', 1, 'false',
    'False', 0.

    Examples:

    >>> to_bool("true")
    True
    >>> to_bool("0")
    False
    >>> to_bool(True)
    True
    """
    if isinstance(s, bool):
        return s
    elif s.lower() in ['true', '1']:
        return True
    elif s.lower() in ['false', '0']:
        return False
    else:
        raise ValueError("Can't cast '%s' to bool" % (s))



#!/usr/bin/env python

# ansible_cmd
#
# Generate host overview (configuration management database) from ansible fact
# gathering output.
#
# Usage:
#
#   $ ansible -m setup --tree out all
#   $ ansible-cmdb out > cmdb.html
#

import optparse
import sys
import os
import logging
from mako import exceptions
import ansiblecmdb
import ansiblecmdb.util as util
import ansiblecmdb.render as render


# Verify Python version
if sys.version_info < (2, 7):
    sys.stderr.write(
        "Ansible-cmdb requires Python v2.7+. You are running {}. Support for "
        "Python v2.6 supported ended October 2013. You should upgrade to a "
        "newer version.\n".format(sys.version))
    sys.exit(1)


def get_logger():
    """
    Instantiate a logger.
    """
    root = logging.getLogger()
    root.setLevel(logging.WARNING)
    ch = logging.StreamHandler(sys.stderr)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(message)s')
    ch.setFormatter(formatter)
    root.addHandler(ch)
    return root


def get_data_dir():
    """
    Find out our installation prefix and data directory. These can be in
    different places depending on how ansible-cmdb was installed.
    """
    data_dir_paths = [
        os.path.join(os.path.dirname(ansiblecmdb.__file__), 'data'),
        os.path.join(os.path.dirname(sys.argv[0]), '..', 'lib', 'ansiblecmdb', 'data'),
        '/usr/local/lib/ansiblecmdb/data',
        '/usr/lib/ansiblecmdb/data',
    ]

    data_dir = util.find_path(data_dir_paths, 'tpl/html_fancy.tpl')
    if not data_dir:
        sys.stdout.write("Couldn't find the data dir for the templates. I tried: {0}\n".format(", ".join(data_dir_paths)))
        sys.exit(1)

    return data_dir


def get_hosts_files(option):
    """
    Find out the location of the `hosts` file. This looks in multiple places
    such as the `-i` option, current dir and ansible configuration files. The
    first match is returned as a list.
    """
    if option is not None:
        return option.split(',')

    # Use hosts file from the current dir if it exists
    if os.path.isfile('hosts'):
        return ['hosts']

    # Perhaps it's configured in a configuration file. Try to find a
    # configuration file and see if it contains a `hostsfile` entry.
    config_locations = [
        '.',
        '/etc/ansible/'
    ]
    config_dir = util.find_path(config_locations, 'ansible.cfg')
    log.debug('config_dir = {0}'.format(config_dir))
    if config_dir:
        with open(os.path.join(config_dir, 'ansible.cfg'), 'r') as cf:
            for line in cf:
                if line.startswith('hostfile'):
                    return [line.split('=', 1)[1].strip()]


def parse_user_params(user_params):
    """
    Parse the user params (-p/--params) and them as a dict.
    """
    if user_params:
        params = {}
        try:
            for param in options.params.split(','):
                param_key, param_value = param.split('=', 1)
                params[param_key] = param_value
        except ValueError as e:
            sys.stdout.write("Invalid params specified. Should be in format: <key=value>[,<key=value>..]\n")
            sys.exit(1)
        return params
    else:
        return {}


if __name__ == "__main__":
    log = get_logger()
    data_dir = get_data_dir()
    tpl_dir = os.path.join(data_dir, 'tpl')
    static_dir = os.path.join(data_dir, 'static')
    version = open(os.path.join(data_dir, 'VERSION')).read().strip()

    parser = optparse.OptionParser(version="%prog v{0}".format(version))
    parser.set_usage(os.path.basename(sys.argv[0]) + " [option] <dir> > output.html")
    parser.add_option("-t", "--template", dest="template", action="store", default='html_fancy', help="Template to use. Default is 'html_fancy'")
    parser.add_option("-i", "--inventory", dest="inventory", action="store", default=None, help="Inventory to read extra info from")
    parser.add_option("-f", "--fact-cache", dest="fact_cache", action="store_true", default=False, help="<dir> contains fact-cache files")
    parser.add_option("-p", "--params", dest="params", action="store", default=None, help="Params to send to template")
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False, help="Show debug output")
    parser.add_option("-q", "--quiet", dest="quiet", action="store_true", default=False, help="Don't report warnings")
    parser.add_option("-c", "--columns", dest="columns", action="store", default=None, help="Show only given columns")
    parser.add_option("--exclude-cols", dest="exclude_columns", action="store", default=None, help="Exclude cols from output")
    (options, args) = parser.parse_args()

    if len(args) < 1:
        parser.print_usage()
        sys.stderr.write("The <dir> argument is mandatory\n")
        sys.exit(1)

    if options.quiet:
        log.setLevel(logging.ERROR)
    elif options.debug:
        log.setLevel(logging.DEBUG)

    hosts_files = get_hosts_files(options.inventory)

    # Handle template params
    params = {
        'lib_dir': data_dir,  # Backwards compatibility for custom templates < ansible-cmdb v1.7
        'data_dir': data_dir,
        'version': version,
        'log': log,
        'columns': None,
        'exclude_columns': None,
    }
    params.update(parse_user_params(options.params))
    if options.columns is not None:
        params['columns'] = options.columns.split(',')
    if options.exclude_columns is not None:
        params['exclude_columns'] = options.exclude_columns.split(',')

    # Log some debug information
    log.debug('data_dir = {0}'.format(data_dir))
    log.debug('tpl_dir = {0}'.format(tpl_dir))
    log.debug('static_dir = {0}'.format(static_dir))
    log.debug('inventory files = {0}'.format(hosts_files))
    log.debug('template params = {0}'.format(params))

    ansible = ansiblecmdb.Ansible(args, hosts_files, options.fact_cache, debug=options.debug)

    # Render a template with the gathered host info
    renderer = render.Render(options.template, ['.', tpl_dir])
    if renderer.tpl_file is None:
        sys.stderr.write("Template '{0}' not found at any of \n  {1}\n".format(options.template, "\n  ".join(renderer.tpl_possibilities)))
        sys.exit(1)

    # Make sure we always output in UTF-8, regardless of the user's locale /
    # terminal encoding. This is different in Python 2 and 3.
    try:
        output = renderer.render(ansible.hosts, params)
        if output:
            if sys.version_info[0] == 3:
                sys.stdout.buffer.write(output.lstrip())
            else:
                sys.stdout.write(output.lstrip())
    except Exception as err:
        full_err = exceptions.text_error_template().render().replace("\n", "\n    ")
        debug_cmd = "{0} -d {1}".format(sys.argv[0], ' '.join(sys.argv[1:]))
        debug_txt = ("Whoops, it looks like something went wrong while rendering the template.\n\n"
                     "The reported error was: {0}: {1}\n\nThe full error was:{2}\n"
                     "The output is probably not correct.\n\n".format(err.__class__.__name__, err, full_err))
        if err.__class__.__name__ == "KeyError":
            debug_txt += "!!! This is probably a problem with missing information in your host facts\n\n"
        if not options.debug:
            debug_txt += ("You can re-run ansible-cmdb with the -d switch to turn on debugging\n"
                          "to get an insight in what might be going wrong:\n\n"
                          "  {0}\n\n".format(debug_cmd))
        debug_txt += \
"""\
You can report a bug on the issue tracker:

  https://github.com/fboender/ansible-cmdb/issues

Please include the debugging output (-d switch) in the report!

If you can, also include the hosts file and the facts file for the last host
that rendered properly ('Rendering host...' in the output. If these files must
remain confidential, you can send them to ferry.boender@gmail.com instead.
"""
        sys.stderr.write(debug_txt)
        sys.exit(1)



#!/usr/bin/env python
import os
import sys
import re
from distutils.core import setup
from setuptools import find_packages

def get_long_description():
    path = os.path.join(os.path.dirname(__file__), 'README.md')
    with open(path) as f:
        return f.read()

def get_version():
    return open('src/ansiblecmdb/data/VERSION', 'r').read().strip()

def get_data_files(path, strip='', prefix=''):
    data_files = []
    for dirpath, dirnames, filenames in os.walk(path):
        files = [os.path.join(dirpath, filename) for filename in filenames]
        data_files.append( [prefix + dirpath[len(strip):], files] )
    return data_files


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    print('You should also add a git tag for this version:')
    print(' git tag {0}'.format(get_version()))
    print(' git push --tags')
    sys.exit()

setup(
    name='ansible-cmdb',
    version=get_version(),
    license='MIT',
    description='Generate host overview from ansible fact gathering output',
    long_description=get_long_description(),
    url='https://github.com/fboender/ansible-cmdb',

    author='Ferry Boender',
    author_email='ferry.boender@electricmonk.nl',

    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    data_files=\
        get_data_files(
            'src/ansiblecmdb/data',
            strip='src',
            prefix='lib'
        ) +
        [['lib/ansiblecmdb/', ['src/ansible-cmdb.py']]],
    zip_safe=False,
    install_requires=['mako', 'pyyaml', 'ushlex', 'jsonxs'],
    scripts=[
        'src/ansible-cmdb',
    ],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
)



"""
Custom Yaml library wrapper, because Yaml is a bag of shit.
"""

try:
    import yaml
except ImportError:
    import yaml3 as yaml

class StupidYAMLShit(yaml.SafeLoader):
    """
    FUCK PYYAML. This class overrides some insanely deep shit which took at
    least two hours to get working. This class overrides SafeLoader and handles
    tags (e.g. '!bullshit') nonsense in PyYAML, because obviously there is no
    ignore_tags option or a simple callback that actually works. That would be
    user-friendly, and user-friendliness is insanity, amirite?!

    Also, there is no single entry point to hook into this, so we need to
    specifically inherit from *SafeLoader* and not from *Loader*. Thanks for
    wasting some more of my fucking time PyYAML, you turd.

    On a pyyaml-rage-induced side note: How many apps are vulnerable because
    all the PyYaml docs mention 'load' and not 'safe_load'? Was this diseased
    pile of gunk written by a PHP programmer?!

    """
    def handle_tag(self, node_name, node):
        # I just *know* there are gonna be problems with simply returning a
        # Scalar, but I don't give a fuck at this point.
        if node_name == "vault":
            new_node = yaml.ScalarNode(node_name, 'ENCRYPTED CONTENTS REDACTED')
        else:
            new_node = yaml.ScalarNode(node_name, node.value)

        return self.construct_scalar(new_node)


# Fugly!
StupidYAMLShit.add_multi_constructor(
    u'!',
    StupidYAMLShit.handle_tag)


def safe_load(contents):
    return yaml.load(contents, Loader=StupidYAMLShit)



import sys
import re
import json
import traceback
import logging
if sys.version_info[0] == 2:
    # Python 2.x shlex doesn't support unicode
    import ushlex as shlex
else:
    import shlex
from . import ihateyaml


class HostsParser(object):
    """
    Parse Ansible hosts inventory contents and generate a dictionary of hosts
    (self.hosts) with the information. Each host will have all the groups that
    it belongs to (including parent groups) and all the variables defined for
    it (including variables from parent groups and 'vars' groups).
    """
    def __init__(self, hosts_contents):
        self.hosts_contents = hosts_contents
        self.hosts = {}
        self.log = logging.getLogger(__name__)

        # Get a list of host, children and vars sections in the hosts file
        try:
            self.sections = self._parse_hosts_contents(hosts_contents)

            # Initialize each unique host found in the hosts file
            for hostname in self._get_distinct_hostnames():
                self.hosts[hostname] = {
                    'groups': set(),
                    'hostvars': {}
                }

            # Go through the different types of sections and apply each section to
            # the hosts it covers. 'Applying a section' means adding the group name
            # and variables defined in the section to each host the section covers.
            for section in filter(lambda s: s['type'] == 'children', self.sections):
                self._apply_section(section, self.hosts)
            for section in filter(lambda s: s['type'] == 'vars', self.sections):
                self._apply_section(section, self.hosts)
            for section in filter(lambda s: s['type'] == 'hosts', self.sections):
                self._apply_section(section, self.hosts)
        except ValueError:
            tb = traceback.format_exc()
            self.log.warn("Error while parsing hosts contents: '{0}'. Invalid hosts file?".format(tb))

    def _parse_hosts_contents(self, hosts_contents):
        """
        Parse the inventory contents. This returns a list of sections found in
        the inventory, which can then be used to figure out which hosts belong
        to which groups and such. Each section has a name, a type ('hosts',
        'children', 'vars') and a list of entries for that section. Entries
        consist of a hostname and the variables. For 'vars' sections, the
        hostname is None.

        For example:
            [production:children]
            frontend  purpose="web"
            db        purpose="db"
        Returns:
            {
                'name': 'production',
                'type': 'children',
                'entries': [
                    {'name': 'frontend', 'hostvars': {'purpose': 'web'}},
                    {'name': 'db', 'hostvars': {'purpose': 'db'}},
                ]
            }
        """
        sections = []
        cur_section = {
            'type': 'hosts',
            'name': None,
            'entries': []
        }

        for line in hosts_contents:
            line = line.strip()
            if line.startswith('#') or not line:
                continue
            elif line.startswith('['):
                sections.append(cur_section)
                section_type, name = self._parse_line_section(line)
                cur_section = {
                    'type': section_type,
                    'name': name,
                    'entries': []
                }
            else:
                name, vars = self._parse_line_entry(line, cur_section['type'])
                entry = {
                    'name': name,
                    'hostvars': vars
                }
                cur_section['entries'].append(entry)
        sections.append(cur_section)
        return sections

    def _parse_line_section(self, line):
        """
        Parse a line containing a group definition. Returns a tuple:
        (group_type, group_name), where group_type is in the set ('hosts',
        'children', 'vars').

        For example:
            [prod]
        Returns:
            ('hosts', 'prod')

        For example:
            [prod:children]
        Returns:
            ('children', 'prod')
        """
        m = re.match("\[(.*)\]", line)
        group_def = m.groups()[0]
        if ':' in group_def:
            group_name, group_type = group_def.split(':')
        else:
            group_name = group_def
            group_type = 'hosts'

        return (group_type, group_name)

    def _parse_line_entry(self, line, type):
        """
        Parse a section entry line into its components. In case of a 'vars'
        section, the first field will be None. Otherwise, the first field will
        be the unexpanded host or group name the variables apply to.

        For example:
            [production:children]
            frontend  purpose="web"    # The line we process
        Returns:
            ('frontend', {'purpose': 'web'})

        For example:
            [production:vars]
            purpose="web"              # The line we process
        Returns:
            (None, {'purpose': 'web'})

        Undocumented feature:
            [prod:vars]
            json_like_vars=[{'name': 'htpasswd_auth'}]
        Returns:
            (None, {'name': 'htpasswd_auth'})
        """

        name = None
        key_values = {}

        if type == 'vars':
            key_values = self._parse_line_vars(line)
        else:
            tokens = shlex.split(line.strip())
            name = tokens.pop(0)
            try:
                key_values = self._parse_vars(tokens)
            except ValueError:
                self.log.warning("Unsupported vars syntax. Skipping line: {0}".format(line))
                return (name, {})
        return (name, key_values)

    def _parse_line_vars(self, line):
        """
        Parse a line in a [XXXXX:vars] section.
        """
        key_values = {}

        # Undocumented feature allows json in vars sections like so:
        #   [prod:vars]
        #   json_like_vars=[{'name': 'htpasswd_auth'}]
        # We'll try this first. If it fails, we'll fall back to normal var
        # lines. Since it's undocumented, we just assume some things.
        k, v = line.strip().split('=', 1)
        if v.startswith('['):
            try:
                list_res = ihateyaml.safe_load(v)
                if isinstance(list_res[0], dict):
                    key_values = list_res[0]
                    return key_values
            except ValueError:
                pass

        # Guess it's not YAML. Parse as normal host variables
        tokens = shlex.split(line.strip())
        key_values = self._parse_vars(tokens)
        return key_values

    def _parse_vars(self, tokens):
        """
        Given an iterable of tokens, returns variables and their values as a
        dictionary.

        For example:
            ['dtap=prod', 'comment=some comment']
        Returns:
            {'dtap': 'prod', 'comment': 'some comment'}
        """
        key_values = {}
        for token in tokens:
            if token.startswith('#'):
                # End parsing if we encounter a comment, which lasts
                # until the end of the line.
                break
            else:
                k, v = token.split('=', 1)
                key = k.strip()
                key_values[key] = v.strip()
        return key_values

    def _get_distinct_hostnames(self):
        """
        Return a set of distinct hostnames found in the entire inventory.
        """
        hostnames = []
        for section in self.sections:
            hostnames.extend(self._group_get_hostnames(section['name']))
        return set(hostnames)

    def _apply_section(self, section, hosts):
        """
        Recursively find all the hosts that belong in or under a section and
        add the section's group name and variables to every host.
        """
        # Add the current group name to each host that this section covers.
        if section['name'] is not None:
            for hostname in self._group_get_hostnames(section['name']):
                hosts[hostname]['groups'].add(section['name'])

        # Apply variables
        func_map = {
            "hosts": self._apply_section_hosts,
            "children": self._apply_section_children,
            "vars": self._apply_section_vars,
        }
        func = func_map[section['type']]
        func(section, hosts)

    def _apply_section_hosts(self, section, hosts):
        """
        Add the variables for each entry in a 'hosts' section to the hosts
        belonging to that entry.
        """
        for entry in section['entries']:
            for hostname in self.expand_hostdef(entry['name']):
                host = hosts[hostname]
                for var_key, var_val in entry['hostvars'].items():
                    host['hostvars'][var_key] = var_val

    def _apply_section_children(self, section, hosts):
        """
        Add the variables for each entry in a 'children' section to the hosts
        belonging to that entry.
        """
        for entry in section['entries']:
            for hostname in self._group_get_hostnames(entry['name']):
                host = hosts[hostname]
                for var_key, var_val in entry['hostvars'].items():
                    host['hostvars'][var_key] = var_val

    def _apply_section_vars(self, section, hosts):
        """
        Apply the variables in a 'vars' section to each host belonging to the
        group the section refers to.
        """
        for hostname in self._group_get_hostnames(section['name']):
            host = hosts[hostname]
            for entry in section['entries']:
                for var_key, var_val in entry['hostvars'].items():
                    host['hostvars'][var_key] = var_val

    def _group_get_hostnames(self, group_name):
        """
        Recursively fetch a list of each unique hostname that belongs in or
        under the group. This includes hosts in children groups.
        """
        hostnames = []

        hosts_section = self._get_section(group_name, 'hosts')
        if hosts_section:
            for entry in hosts_section['entries']:
                hostnames.extend(self.expand_hostdef(entry['name']))

        children_section = self._get_section(group_name, 'children')
        if children_section:
            for entry in children_section['entries']:
                hostnames.extend(self._group_get_hostnames(entry['name']))

        return hostnames

    def _get_section(self, name, type):
        """
        Find and return a section with `name` and `type`
        """
        for section in self.sections:
            if section['name'] == name and section['type'] == type:
                return section
        return None

    def expand_hostdef(self, hostdef):
        """
        Expand a host definition (e.g. "foo[001:010].bar.com") into seperate
        hostnames. Supports zero-padding, numbered ranges and alphabetical
        ranges. Multiple patterns in a host defnition are also supported.
        Returns a list of the fully expanded hostnames. Ports are also removed
        from hostnames as a bonus (e.g. "foo.bar.com:8022" -> "foo.bar.com")
        """
        try:
            hosts_todo = [hostdef]
            hosts_done = []

            # Keep going through the todo list of hosts until they no longer have a
            # pattern in them. We only handle the first pattern found in the host for
            # each iteration of the while loop. If more patterns are present, the
            # partially expanded host(s) gets added back to the todo list.
            while hosts_todo:
                host = hosts_todo.pop(0)
                if '[' not in host:
                    hosts_done.append(host)
                    continue

                # Extract the head, first pattern and tail. E.g. foo[0:3].bar.com ->
                # head="foo", pattern="0:3", tail=".bar.com"
                head, rest = host.split('[', 1)
                pattern, tail = rest.split(']', 1)
                start, end = pattern.split(':')
                fill = False
                if start.startswith('0') and len(start) > 0:
                    fill = len(start)

                try:
                    for i in range(int(start), int(end) + 1):
                        if fill:
                            range_nr = str(i).zfill(fill)
                        else:
                            range_nr = i
                        new_host = '{0}{1}{2}'.format(head, range_nr, tail)
                        if '[' in new_host:
                            hosts_todo.append(new_host)
                        else:
                            hosts_done.append(new_host)
                except ValueError:
                    for i in range(ord(start), ord(end) + 1):
                        new_host = '{0}{1}{2}'.format(head, chr(i), tail)
                        if '[' in new_host:
                            hosts_todo.append(new_host)
                        else:
                            hosts_done.append(new_host)

            # Strip port numbers off and return
            return [host_name.split(':')[0] for host_name in hosts_done]
        except Exception as e:
            self.log.warning("Couldn't parse host definition '{0}': {1}".format(hostdef, e))
            return []


class DynInvParser(object):
    """
    Parse output of a dyanmic inventory script.
    """
    def __init__(self, dynvinv_contents):
        self.dynvinv_contents = dynvinv_contents
        self.hosts = {}
        self.dynvinv_json = json.loads(self.dynvinv_contents)
        self.log = logging.getLogger(__name__)

        for k, v in self.dynvinv_json.items():
            if k.startswith('_meta'):
                # Meta member contains hostvars
                self._parse_meta(v)
            elif k.startswith('_'):
                # Some unknown private member we don't support
                pass
            elif k == "all":
                # The 'all' group is not interesting for ansible-cmdb
                pass
            else:
                # Group definition
                self._parse_group(k, v)

    def _get_host(self, hostname):
        """
        Get an existing host or otherwise initialize a new empty one.
        """
        if hostname not in self.hosts:
            self.hosts[hostname] = {
                'groups': set(),
                'hostvars': {}
            }
        return self.hosts[hostname]

    def _parse_group(self, group_name, group):
        """
        Parse a group definition from a dynamic inventory. These are top-level
        elements which are not '_meta(data)'.
        """
        if type(group) == dict:
            # Example:
            #     {
            #       "mgmt": {
            #         "hosts": [ "mgmt01", "mgmt02" ],
            #         "vars": {
            #           "eth0": {
            #             "onboot": "yes",
            #             "nm_controlled": "no"
            #           }
            #         }
            #       }
            #     }
            #
            hostnames_in_group = set()

            # Group member with hosts and variable definitions.
            for hostname in group.get('hosts', []):
                self._get_host(hostname)['groups'].add(group_name)
                hostnames_in_group.add(hostname)
            # Apply variables to all hosts in group
            for var_key, var_val in group.get('vars', {}).items():
                for hostname in hostnames_in_group:
                    self._get_host(hostname)['hostvars'][var_key] = var_val
        elif type(group) == list:
            # List of hostnames for this group
            for hostname in group:
                self._get_host(hostname)['groups'].add(group_name)
        else:
            self.log.warning("Invalid element found in dynamic inventory output: {0}".format(type(group)))

    def _parse_meta(self, meta):
        """
        Parse the _meta element from a dynamic host inventory output.
        """
        for hostname, hostvars in meta.get('hostvars', {}).items():
            for var_key, var_val in hostvars.items():
                self._get_host(hostname)['hostvars'][var_key] = var_val
